@@ -9,6 +9,8 @@ from enum import Enum
 
 class ModelType(Enum):
     XRAY = 1
+    KDD = 10
+    MRI = 11
     CREDIT = 2
     MNIST = 3
     CIFAR = 4
@@ -16,7 +18,8 @@ class ModelType(Enum):
 
 class SamplingMode(Enum):
     BASELINE = 1
-    GREEDY = 2
+    IMPACT = 2
+    # STOCHASTIC is working in progress now
     STOCHASTIC = 3
     SCALE = 0
     OTHER = -1
